/* id3tag.c - 2001/02/16 */
/*
 *  EasyTAG - Viewer/Editor for MP3 Tags
 *  Copyright (C) 2001  Jerome Couderc <j.couderc@ifrance.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * iconv support
 * (C)Copyright 2001 by Hiroshi Takekawa <sian@big.or.jp>
 *
 * USE_ICONV: When defined, convert tag characters from FILE_CODE to
 * SYSTEM_CODE in reading, and convert backward in writing.
 */

#define USE_ICONV
/* Should be configurable? */
#ifdef USE_ICONV
#  define FILE_CODE "SJIS"
#  define SYSTEM_CODE "EUC-JP"
#  define ICONV_OUTPUT_SIZE 65536
#endif

#include <id3.h>
#include <gtk/gtk.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#ifdef USE_ICONV
#  include <iconv.h>
#endif

#include "id3tag.h"
#include "genres.h"
#include "easytag.h"
#include "mylocale.h"
#include "dirbrowser.h"
#include "config.h"
#include "undo.h"
#include "misc.h"


/****************
 * Declarations *
 ****************/


/**************
 * Prototypes *
 **************/
void Id3tag_Initialize_Tag (MP3_File *mp3file);
ID3_C_EXPORT size_t my_ID3Field_GetASCII(const ID3Field *field,char *buffer,size_t maxChars,index_t itemNum);


/*************
 * Functions *
 *************/

#ifdef USE_ICONV

static void
convert_string(unsigned char **sp, char *from, char *to)
{
  iconv_t cd;
  char d[ICONV_OUTPUT_SIZE];
  char *s = (char *)*sp;
  char *inptr, *outptr;
  size_t avail, insize, nconv;
  static int error_reported = 0;

  if (!s)
    return;
  if ((insize = strlen(s)) == 0)
    return;

  if ((cd = iconv_open(to, from)) == (iconv_t)-1) {
    if (errno == EINVAL) {
      if (!error_reported) {
	fprintf(stderr, "Conversion from `" FILE_CODE "' to `" SYSTEM_CODE "' is no available.\n");
	error_reported++;
      }
      return;
    }
  }

  inptr = s;
  outptr = d;
  avail = ICONV_OUTPUT_SIZE - 1;

  if ((nconv = iconv(cd, &inptr, &insize, &outptr, &avail)) != (size_t)-1) {
    *outptr = '\0';
    if (strlen(s) >= outptr - d) {
      strcpy(s, d);
    } else {
      g_free(*sp);
      *sp = g_strdup(d);
    }
  } else {
    switch (errno) {
    case E2BIG:
      fprintf(stderr, "Increase ICONV_OUTPUT_SIZE and recompile.\n");
      break;
    case EILSEQ:
      fprintf(stderr, "Invalid sequence passed.\n");
      break;
    case EINVAL:
      fprintf(stderr, "Incomplete multi-byte sequence passed.\n");
      break;
    default:
      perror("convert_string");
      break;
    }
  }

  iconv_close(cd);

  return;
}

static void
convert_file2system(unsigned char **sp)
{
  convert_string(sp, FILE_CODE, SYSTEM_CODE);
}

static void
convert_system2file(unsigned char **sp)
{
  convert_string(sp, SYSTEM_CODE, FILE_CODE);
}

#endif /* USE_ICONV */

/* 
 * Check the extension of the file
 * Return 1 if it's an MP3; else 0.
 */
int Check_If_Mp3_By_Extension (gchar *filename)
{
	FILE *file;
	gchar *extension = NULL;

	if ( (file=fopen(filename,"r"))==NULL )
	{
		g_print(_("ERROR while opening file: '%s' (%s)\n"),filename,g_strerror(errno));
		return 0;
	}
	fclose(file);

	extension = strrchr(filename,'.');
	if (extension && !strcasecmp(extension,".mp3"))
		return 1;
	else
		return 0;
}


/*
 * Set to NULL all tag infos into the MP3_File structure
 */
void Id3tag_Initialize_Tag (MP3_File *mp3file)
{
	mp3file->id3.title    = NULL;
	mp3file->id3.artist   = NULL;
	mp3file->id3.album    = NULL;
	mp3file->id3.year     = NULL;
	mp3file->id3.track    = NULL;
	mp3file->id3.genre    = NULL;
	mp3file->id3.comment  = NULL;
}


/*
 * Read id3v1.x / id3v2 tag and load data into the MP3_File structure using id3lib functions.
 * Returns 1 if succeed.
 * If a tag entry exists (ex: title), we allocate memory, else stay to NULL
 */
int Id3tag_Read_Tag (MP3_File *mp3file)
{
	FILE *file;
	ID3Tag *id3_tag = NULL;	/* Tag defined by the id3lib */
	gint Item_Num;


	/* Init the tag structure */
	Id3tag_Initialize_Tag (mp3file);

	if ( (file=fopen(mp3file->file,"r"))==NULL )
	{
		g_print(_("ERROR while opening file: '%s' (%s)\n\a"),mp3file->file,g_strerror(errno));
		return 0;
	}
	fclose(file);

	/* Note : The last value of ID3Field_GetASCII function represents the number of
	 * an item (we can have one or two or more title defined in the tag).
	 * with id3lib-3.7.12 and previous, the first item is '1'
	 * with id3lib-3.7.13 and later, the first item is '0'
	 */
	if (   (ID3LIB_MAJOR_VERSION < 3)
	    || (ID3LIB_MAJOR_VERSION == 3 && ID3LIB_MINOR_VERSION < 7)
	    || (ID3LIB_MAJOR_VERSION == 3 && ID3LIB_MINOR_VERSION == 7 && ID3LIB_PATCH_VERSION < 13) )
        Item_Num = 1;
    else
        Item_Num = 0;


    /* Get data from tag */
	if ( (id3_tag = ID3Tag_New()) )
	{
		ID3Frame *id3_frame;
		ID3Field *id3_field;
		luint frm_size;

		/* Link the file to the tag */
		frm_size = ID3Tag_Link(id3_tag, mp3file->file);


		/*********
		 * Title *
		 *********/
		if ( (id3_frame = ID3Tag_FindFrameWithID(id3_tag,ID3FID_TITLE)) )
		{
			if ( (id3_field = ID3Frame_GetField(id3_frame,ID3FN_TEXT)) )
			{
				luint num_chars;
				mp3file->id3.title = g_malloc0(MAX_STRING_LEN+1);
				num_chars = my_ID3Field_GetASCII(id3_field,mp3file->id3.title,MAX_STRING_LEN,Item_Num);
				Id3tag_Strip(mp3file->id3.title);
			}
    	}


		/**********
		 * Artist *
		 **********/
		if ( (id3_frame = ID3Tag_FindFrameWithID(id3_tag,ID3FID_LEADARTIST)) )
		{
			if ( (id3_field = ID3Frame_GetField(id3_frame,ID3FN_TEXT)) )
			{
				luint num_chars;
				mp3file->id3.artist = g_malloc0(MAX_STRING_LEN+1);
				num_chars = my_ID3Field_GetASCII(id3_field,mp3file->id3.artist,MAX_STRING_LEN,Item_Num);
				Id3tag_Strip(mp3file->id3.artist);
			}
		}


		/*********
		 * Album *
		 *********/
		if ( (id3_frame = ID3Tag_FindFrameWithID(id3_tag,ID3FID_ALBUM)) )
		{
			if ( (id3_field = ID3Frame_GetField(id3_frame,ID3FN_TEXT)) )
			{
				luint num_chars;
				mp3file->id3.album = g_malloc0(MAX_STRING_LEN+1);
				num_chars = my_ID3Field_GetASCII(id3_field,mp3file->id3.album,MAX_STRING_LEN,Item_Num);
				Id3tag_Strip(mp3file->id3.album);
			}
		}


		/********
		 * Year *
		 ********/
		if ( (id3_frame = ID3Tag_FindFrameWithID(id3_tag,ID3FID_YEAR)) )
		{
			if ( (id3_field = ID3Frame_GetField(id3_frame,ID3FN_TEXT)) )
			{
				luint num_chars;
				mp3file->id3.year = g_malloc0(MAX_STRING_LEN+1);
				num_chars = my_ID3Field_GetASCII(id3_field,mp3file->id3.year,MAX_STRING_LEN,Item_Num);
				Id3tag_Strip(mp3file->id3.year);
			}
		}


		/*********
		 * Track *
		 *********/
		if ( (id3_frame = ID3Tag_FindFrameWithID(id3_tag,ID3FID_TRACKNUM)) )
		{
			if ( (id3_field = ID3Frame_GetField(id3_frame,ID3FN_TEXT)) )
			{
				luint num_chars;
				gchar buffer[MAX_STRING_LEN+1];
				mp3file->id3.track = g_malloc0(MAX_STRING_LEN+1);
				num_chars = my_ID3Field_GetASCII(id3_field,buffer,MAX_STRING_LEN,Item_Num);
				// Just to have numbers like '01', '05', '12', ...
				snprintf(mp3file->id3.track,MAX_STRING_LEN,"%.2d",atoi(buffer));
				Id3tag_Strip(mp3file->id3.track);
			}
		}


		/*********
		 * Genre *
		 *********/
		if ( (id3_frame = ID3Tag_FindFrameWithID(id3_tag,ID3FID_CONTENTTYPE)) )
		{
			if ( (id3_field = ID3Frame_GetField(id3_frame,ID3FN_TEXT)) )
			{
				luint num_chars;
				mp3file->id3.genre = g_malloc0(MAX_STRING_LEN+1);
				num_chars = my_ID3Field_GetASCII(id3_field,mp3file->id3.genre,MAX_STRING_LEN,Item_Num);
				Id3tag_Strip(mp3file->id3.genre);

				if ( mp3file->id3.genre[0]=='(' && strchr(mp3file->id3.genre,')') )
				{
					/* Convert a genre written as '(3)Dance' into '(3)' */
					gint i;
					for (i=1;mp3file->id3.genre[i]!=')';i++) {}
					mp3file->id3.genre[++i] = '\0';
				}
				else //if(Id3tag_String_To_Genre(mp3file->id3.genre)!=0xFF)
				{
					/* Convert a genre written as 'Dance' into '(3)' */
					/* If the genre is unknown 'my_perso_genre', returns (255) */
					gchar *buf = g_strdup_printf("(%d)",Id3tag_String_To_Genre(mp3file->id3.genre));
					strncpy(mp3file->id3.genre,buf,MAX_STRING_LEN);
					g_free(buf);
				}
			}
		}


		/***********
		 * Comment *
		 ***********/
		if ( (id3_frame = ID3Tag_FindFrameWithID(id3_tag,ID3FID_COMMENT)) )
		{
			if ( (id3_field = ID3Frame_GetField(id3_frame,ID3FN_TEXT)) )
			{
				luint num_chars;
				mp3file->id3.comment = g_malloc0(MAX_STRING_LEN+1);
				num_chars = my_ID3Field_GetASCII(id3_field,mp3file->id3.comment,MAX_STRING_LEN,Item_Num);
				Id3tag_Strip(mp3file->id3.comment);
			}
			/*if ( (id3_field = ID3Frame_GetField(id3_frame,ID3FN_DESCRIPTION)) )
			{
				luint num_chars;
				gchar *comment1 = g_malloc0(MAX_STRING_LEN+1);
				num_chars = ID3Field_GetASCII(id3_field,comment1,MAX_STRING_LEN,Item_Num);
				g_free(comment1);
			}
			if ( (id3_field = ID3Frame_GetField(id3_frame,ID3FN_LANGUAGE)) )
			{
				luint num_chars;
				gchar *comment2 = g_malloc0(MAX_STRING_LEN+1);
				num_chars = ID3Field_GetASCII(id3_field,comment2,MAX_STRING_LEN,Item_Num);
				g_free(comment2);
			}
			*/
		}

        /* Displays for debugging */
/*        g_print("\n### '%s' ###:\n",g_basename(mp3file->file));
        g_print("nbr frames: %d\n",ID3Tag_NumFrames(id3_tag));
        g_print("id3v1:%d - id3v2:%d\n",ID3Tag_HasTagType(id3_tag,ID3TT_ID3V1),ID3Tag_HasTagType(id3_tag,ID3TT_ID3V2));
    	g_print("Title  : '%s'\n",mp3file->id3.title);
    	g_print("Artist : '%s'\n",mp3file->id3.artist);
    	g_print("Album  : '%s'\n",mp3file->id3.album);
    	g_print("Year   : '%s'\n",mp3file->id3.year);
    	g_print("Track  : '%s'\n",mp3file->id3.track);
    	g_print("Genre  : '%s'\n",mp3file->id3.genre);
    	g_print("Comment: '%s'\n",mp3file->id3.comment);
*/
		/* Free allocated data */
		ID3Tag_Delete(id3_tag);
	}


	if (USE_TRANSLATION_TABLE)
	{
		serv2user((unsigned char *)mp3file->id3.title  );
		serv2user((unsigned char *)mp3file->id3.artist );
		serv2user((unsigned char *)mp3file->id3.album  );
		serv2user((unsigned char *)mp3file->id3.comment);
		serv2user((unsigned char *)mp3file->id3.year   );
		serv2user((unsigned char *)mp3file->id3.track  );
		/* Genre must not to be translated! */
	}

#ifdef USE_ICONV
	convert_file2system((unsigned char **)&mp3file->id3.title  );
	convert_file2system((unsigned char **)&mp3file->id3.artist );
	convert_file2system((unsigned char **)&mp3file->id3.album  );
	convert_file2system((unsigned char **)&mp3file->id3.comment);
	convert_file2system((unsigned char **)&mp3file->id3.year   );
	convert_file2system((unsigned char **)&mp3file->id3.track  );
#endif

	return 1;
}


/*
 * Write the ID3 tags to the file. Returns 1 if succeed, else 0.
 */
int Id3tag_Write_Tag (MP3_File *mp3file)
{
    ID3Tag *id3_tag = NULL;
	gint error = 0;
	gint number_of_frames;
	FILE *file;
    struct stat statbuf;
    mode_t file_permissions = 0;
	gboolean has_title   = 1;
	gboolean has_artist  = 1;
	gboolean has_album   = 1;
	gboolean has_year    = 1;
	gboolean has_track   = 1;
	gboolean has_genre   = 1;
	gboolean has_comment = 1;

	/* Test to know if we can write into the file */
	if ( (file=fopen(mp3file->file,"r+"))==NULL )
	{
		g_print(_("ERROR while opening file: '%s' (%s)\n\a"),mp3file->file,g_strerror(errno));
		return 0;
	}
	fclose(file);

    /* Save permissions of the file (cause they may change with files on NFS) */
    if ( stat(mp3file->file,&statbuf)!=-1 )
        file_permissions = statbuf.st_mode & (S_IRWXU|S_IRWXG|S_IRWXO);

    /* "Untranslate" strings before to build it and to save it */
    if (USE_TRANSLATION_TABLE)
    {
	    user2serv((unsigned char *)mp3file->id3.title  );
	    user2serv((unsigned char *)mp3file->id3.artist );
	    user2serv((unsigned char *)mp3file->id3.album  );
	    user2serv((unsigned char *)mp3file->id3.comment);
	    user2serv((unsigned char *)mp3file->id3.year   );
	    user2serv((unsigned char *)mp3file->id3.track  );
	    /* Genre must not to be translated! */
    }

#ifdef USE_ICONV
    convert_system2file((unsigned char **)&mp3file->id3.title  );
    convert_system2file((unsigned char **)&mp3file->id3.artist );
    convert_system2file((unsigned char **)&mp3file->id3.album  );
    convert_system2file((unsigned char **)&mp3file->id3.comment);
    convert_system2file((unsigned char **)&mp3file->id3.year   );
    convert_system2file((unsigned char **)&mp3file->id3.track  );
#endif

    /* We get again the tag from the file to keep also unused data (by EasyTAG), then
     * we replace the changed data */
    if ( (id3_tag = ID3Tag_New()) )
    {
        ID3Frame *id3_frame;
        ID3Field *id3_field;

		ID3Tag_Link(id3_tag,mp3file->file);


		/*********
		 * Title *
		 *********/
		if (mp3file->id3.title && strlen(mp3file->id3.title)>0 )
		{
			if ( !(id3_frame = ID3Tag_FindFrameWithID(id3_tag,ID3FID_TITLE)) )
			{
				id3_frame = ID3Frame_NewID(ID3FID_TITLE);
				ID3Tag_AttachFrame(id3_tag,id3_frame);
			}

			if ((id3_field = ID3Frame_GetField(id3_frame,ID3FN_TEXT)))
			{
				ID3Field_SetASCII(id3_field,mp3file->id3.title);
			}
		} else
		{
			if ( (id3_frame = ID3Tag_FindFrameWithID(id3_tag,ID3FID_TITLE)) )
				ID3Tag_RemoveFrame(id3_tag,id3_frame);
            has_title = 0;
		}


		/**********
		 * Artist *
		 **********/
		if (mp3file->id3.artist && strlen(mp3file->id3.artist)>0 )
		{
			if ( !(id3_frame = ID3Tag_FindFrameWithID(id3_tag,ID3FID_LEADARTIST)) )
			{
				id3_frame = ID3Frame_NewID(ID3FID_LEADARTIST);
				ID3Tag_AttachFrame(id3_tag,id3_frame);
			}

			if ((id3_field = ID3Frame_GetField(id3_frame,ID3FN_TEXT)))
			{
				ID3Field_SetASCII(id3_field,mp3file->id3.artist);
			}
		} else
		{
			if ( (id3_frame = ID3Tag_FindFrameWithID(id3_tag,ID3FID_LEADARTIST)) )
				ID3Tag_RemoveFrame(id3_tag,id3_frame);
            has_artist = 0;
		}


		/*********
		 * Album *
		 *********/
		if (mp3file->id3.album && strlen(mp3file->id3.album)>0 )
		{
			if ( !(id3_frame = ID3Tag_FindFrameWithID(id3_tag,ID3FID_ALBUM)) )
			{
				id3_frame = ID3Frame_NewID(ID3FID_ALBUM);
				ID3Tag_AttachFrame(id3_tag,id3_frame);
			}

			if ((id3_field = ID3Frame_GetField(id3_frame,ID3FN_TEXT)))
			{
				ID3Field_SetASCII(id3_field,mp3file->id3.album);
			}
		} else
		{
			if ( (id3_frame = ID3Tag_FindFrameWithID(id3_tag,ID3FID_ALBUM)) )
				ID3Tag_RemoveFrame(id3_tag,id3_frame);
            has_album = 0;
		}


		/********
		 * Year *
		 ********/
		if (mp3file->id3.year && strlen(mp3file->id3.year)>0 )
		{
			if ( !(id3_frame = ID3Tag_FindFrameWithID(id3_tag,ID3FID_YEAR)) )
			{
				id3_frame = ID3Frame_NewID(ID3FID_YEAR);
				ID3Tag_AttachFrame(id3_tag,id3_frame);
			}

			if ((id3_field = ID3Frame_GetField(id3_frame,ID3FN_TEXT)))
			{
				ID3Field_SetASCII(id3_field,mp3file->id3.year);
			}
		} else
		{
			if ( (id3_frame = ID3Tag_FindFrameWithID(id3_tag,ID3FID_YEAR)) )
				ID3Tag_RemoveFrame(id3_tag,id3_frame);
            has_year = 0;
		}


		/*********
		 * Track *
		 *********/
		if (mp3file->id3.track && strlen(mp3file->id3.track)>0 && atoi(mp3file->id3.track)!=0)
		{
			if ( !(id3_frame = ID3Tag_FindFrameWithID(id3_tag,ID3FID_TRACKNUM)) )
			{
				id3_frame = ID3Frame_NewID(ID3FID_TRACKNUM);
				ID3Tag_AttachFrame(id3_tag,id3_frame);
			}

			if ((id3_field = ID3Frame_GetField(id3_frame,ID3FN_TEXT)))
			{
				ID3Field_SetASCII(id3_field,mp3file->id3.track);
			}
		} else
		{
			if ( (id3_frame = ID3Tag_FindFrameWithID(id3_tag,ID3FID_TRACKNUM)) )
				ID3Tag_RemoveFrame(id3_tag,id3_frame);
            has_track = 0;
		}


		/*********
		 * Genre *
		 *********/
		if (mp3file->id3.genre && strlen(mp3file->id3.genre)>0 && strcmp(mp3file->id3.genre,"(255)")!=0)
		{
			if ( !(id3_frame = ID3Tag_FindFrameWithID(id3_tag,ID3FID_CONTENTTYPE)) )
			{
				id3_frame = ID3Frame_NewID(ID3FID_CONTENTTYPE);
				ID3Tag_AttachFrame(id3_tag,id3_frame);
			}

			if ((id3_field = ID3Frame_GetField(id3_frame,ID3FN_TEXT)))
			{
				ID3Field_SetASCII(id3_field,mp3file->id3.genre);
			}
		} else
		{
			if ( (id3_frame = ID3Tag_FindFrameWithID(id3_tag,ID3FID_CONTENTTYPE)) )
				ID3Tag_RemoveFrame(id3_tag,id3_frame);
            has_genre = 0;
		}


		/***********
		 * Comment *
		 ***********/
		if (mp3file->id3.comment && strlen(mp3file->id3.comment)>0 )
		{
			if ( !(id3_frame = ID3Tag_FindFrameWithID(id3_tag,ID3FID_COMMENT)) )
			{
				id3_frame = ID3Frame_NewID(ID3FID_COMMENT);
				ID3Tag_AttachFrame(id3_tag,id3_frame);
			}

			if ((id3_field = ID3Frame_GetField(id3_frame,ID3FN_TEXT)))
			{
				ID3Field_SetASCII(id3_field,mp3file->id3.comment);
			}
			/* These 2 following fields allow synchronisation between id3v2 and id3v1 tags with id3lib */
			if ((id3_field = ID3Frame_GetField(id3_frame,ID3FN_DESCRIPTION)))
			{
				ID3Field_SetASCII(id3_field,"ID3v1 Comment");
			}
			if ((id3_field = ID3Frame_GetField(id3_frame,ID3FN_LANGUAGE)))
			{
				ID3Field_SetASCII(id3_field,"XXX");
			}
		} else
		{
			if ( (id3_frame = ID3Tag_FindFrameWithID(id3_tag,ID3FID_COMMENT)) )
				ID3Tag_RemoveFrame(id3_tag,id3_frame);
            has_comment = 0;
		}


		/*********************************
		 * Update id3v1.x and id3v2 tags *
		 *********************************/
		 /* Get the number of frames into the tag, cause if it is
		  * equal to 0, id3lib-3.7.12 doesn't update the tag */
		 number_of_frames = ID3Tag_NumFrames(id3_tag);

        /* If all fields are empty and option STRIP_TAG_WHEN_EMPTY_FIELDS is set to 1
         * we strip the ID3v1.x and ID3v2 tags. Else let see... :)
         */
        if ( STRIP_TAG_WHEN_EMPTY_FIELDS && !has_title && !has_artist && !has_album
             && !has_year && !has_track && !has_genre && !has_comment )
        {
            ID3Tag_Strip(id3_tag,ID3TT_ID3V1);
            ID3Tag_Strip(id3_tag,ID3TT_ID3V2);
        }else
        {
            /* It's better to remove the id3v1 tag before, to synchronize it with the
             * id3v2 tag (else id3lib doesn't do it correctly)
             */
            ID3Tag_Strip(id3_tag,ID3TT_ID3V1);

            /* ID3v1 tag */
            if (WRITE_ID3V1_TAG && number_of_frames!=0)
                ID3Tag_UpdateByTagType(id3_tag,ID3TT_ID3V1);
            else
                ID3Tag_Strip(id3_tag,ID3TT_ID3V1);

            /* ID3v2 tag */
            if (WRITE_ID3V2_TAG && number_of_frames!=0)
                ID3Tag_UpdateByTagType(id3_tag,ID3TT_ID3V2);
            else
                ID3Tag_Strip(id3_tag,ID3TT_ID3V2);
        }

/* FIX ME !!! : Manage error messages !
 * Var: 'error' isn't used
 */

		/* Free allocated data */
		ID3Tag_Delete(id3_tag);
	}

    /* Set permissions for the file, saved before */
    if ( stat(mp3file->file,&statbuf)!=-1 )
        chmod(mp3file->file,file_permissions);

	/* "Retranslate" strings cause we can need them after saving the tag */
	if (USE_TRANSLATION_TABLE)
	{
		serv2user((unsigned char *)mp3file->id3.title  );
		serv2user((unsigned char *)mp3file->id3.artist );
		serv2user((unsigned char *)mp3file->id3.album  );
		serv2user((unsigned char *)mp3file->id3.comment);
		serv2user((unsigned char *)mp3file->id3.year   );
		serv2user((unsigned char *)mp3file->id3.track  );
		/* Genre must not to be translated! */
	}


	if (error)
		g_print(_("Error, tag not written (file: %s)\n"),g_basename(mp3file->file));
	else
	{
		/* Initialize 'changed' indicator: file is now saved, so it's the new reference */
		mp3file->changed &= ~TAG_CHANGED;
		g_print(_("Written tag of '%s'\n"),g_basename(mp3file->file));
	}

	return 1;
}

/**
 ** Predefined id3lib error types.
 **/
/*ID3_ENUM(ID3_Err)
{
  ID3E_NoError = 0,             //< No error reported 				0
  ID3E_NoMemory,                //< No available memory   			1
  ID3E_NoData,                  //< No data to parse 		 		2
  ID3E_BadData,                 //< Improperly formatted data 		3
  ID3E_NoBuffer,                //< No buffer to write to 			4
  ID3E_SmallBuffer,             //< Buffer is too small   			5
  ID3E_InvalidFrameID,          //< Invalid frame id 			 	6
  ID3E_FieldNotFound,           //< Requested field not found	 	7
  ID3E_UnknownFieldType,        //< Unknown field type 				8
  ID3E_TagAlreadyAttached,      //< Tag is already attached to a file	9
  ID3E_InvalidTagVersion,       //< Invalid tag version 				10
  ID3E_NoFile,                  //< No file to parse 			   	11
  ID3E_ReadOnly,                //< Attempting to write to a read-only file   12
  ID3E_zlibError                /< Error in compression/uncompression 		13
};*/



/*
 * Display all tag infos into entry boxes of the user interface : tags fields + filename
 */
void Id3tag_Display_Tag (MP3_File *mp3file)
{
	FILE *file;
	gchar *text;


	if (!FileEntry) return;

	/* Show/hide 'AccessStatusIcon' */
	if ( (file=fopen(mp3file->file,"r+"))==NULL )
	{
		gtk_widget_show_all(AccessStatusIconBox);
	}else
	{
		gtk_widget_hide(AccessStatusIconBox);
		fclose(file);
	}

	/* Show position of current file */
	text = g_strdup_printf("%d/%d:",ListIndex,ListLength);
	gtk_label_set_text(GTK_LABEL(FileIndex),text);
	g_free(text);

	/* Show filename */
	if (mp3file->file_new)
		gtk_entry_set_text(GTK_ENTRY(FileEntry),g_basename(mp3file->file_new));
	else
		gtk_entry_set_text(GTK_ENTRY(FileEntry),g_basename(mp3file->file));
	// Justify to the left text into FileEntry
	gtk_editable_set_position(GTK_EDITABLE(FileEntry),0);

	/* Set the path to the file into BrowserEntry (dirbrowser) */
	text = g_dirname(mp3file->file);
	Browser_Entry_Set_Text(text);
	g_free(text);
	
	
	/**************
	 * Show title *
	 **************/
	if (mp3file->id3.title)
		gtk_entry_set_text(GTK_ENTRY(TitleEntry),mp3file->id3.title);
	else
		gtk_entry_set_text(GTK_ENTRY(TitleEntry),"");

	/***************
	 * Show artist *
	 ***************/
	if (mp3file->id3.artist)
		gtk_entry_set_text(GTK_ENTRY(ArtistEntry),mp3file->id3.artist);
	else
		gtk_entry_set_text(GTK_ENTRY(ArtistEntry),"");

	/**************
	 * Show album *
	 **************/
	if (mp3file->id3.album)
		gtk_entry_set_text(GTK_ENTRY(AlbumEntry),mp3file->id3.album);
	else
		gtk_entry_set_text(GTK_ENTRY(AlbumEntry),"");

	/*************
	 * Show year *
	 *************/
	if (mp3file->id3.year)
		gtk_entry_set_text(GTK_ENTRY(YearEntry),mp3file->id3.year);
	else
		gtk_entry_set_text(GTK_ENTRY(YearEntry),"");

	/**************
	 * Show track *
	 **************/
	if (mp3file->id3.track)
		gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(TrackEntry)->entry),mp3file->id3.track);
	else
		gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(TrackEntry)->entry),"");

	/**************
	 * Show genre *
	 **************/
	if (mp3file->id3.genre)
	{
		gint i;

		//Id3tag_Strip(mp3file->id3.genre);
		text = g_strdup(mp3file->id3.genre);

		/* Get the value into string '(value)' */
		for (i=0; i<strlen(text); i++)
			text[i] = isdigit(text[i]) ? text[i] : ' ';
		i = atoi(text);

		gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(GenreEntry)->entry),Id3tag_Genre_To_String(i));
		g_free(text);
	} else
		gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(GenreEntry)->entry),"");
	

	/****************
	 * Show comment *
	 ****************/
	if (mp3file->id3.comment)
		gtk_entry_set_text(GTK_ENTRY(CommentEntry),mp3file->id3.comment);
	else
		gtk_entry_set_text(GTK_ENTRY(CommentEntry),"");

	/* Get data for undo */
	Undo_Get_Snap_Shot(mp3file);
}



/*
 * Id3_Save_Displayed_Tag_In_List: Keep the changes of the current displayed tag 
 * into the GList but do not apply changes to the mp3 file. (Get data from user interface)
 */
void Id3tag_Save_Displayed_Tag (MP3_File *mp3file)
{
	gchar *buffer = NULL;
	gchar *buffer1 = NULL;
	gint buflen;

	/* Filename */
	/* # This part shouldn't be in this function cause it's not a part of the tag.
	 * # In future it'll be moved to an other function */
	buffer = gtk_entry_get_text(GTK_ENTRY(FileEntry));
	if (strcmp(g_basename(mp3file->file),buffer)==0)
	{
		/* If buffer is equal to mp3file->file, we don't need to rename this file */
		if (mp3file->file_new) g_free(mp3file->file_new);
		mp3file->file_new = NULL;
		/* As file musn't be renamed (also deactivate in Rename_File function) */
		mp3file->changed &= ~FILENAME_CHANGED;
	}else
	{
		if ( (!mp3file->file_new || strcmp(g_basename(mp3file->file_new),buffer)!=0)
		     && (buffer && strlen(buffer)>0 && strchr(buffer,'/')==NULL) )
		{
			/* Get the new file name */
			gchar *dirname = g_dirname(mp3file->file);
			if (mp3file->file_new) g_free(mp3file->file_new);
			mp3file->file_new = g_strconcat(dirname,"/",
				gtk_entry_get_text(GTK_ENTRY(FileEntry)),NULL);
			g_free(dirname);
			/* Filename was changed, and must be saved */
			mp3file->changed |= FILENAME_CHANGED;
		}
	}
	/* # */

	/*********
	 * Title *
	 *********/
	buffer = gtk_entry_get_text(GTK_ENTRY(TitleEntry));
	Id3tag_Strip(buffer);

	if ( (buflen=strlen(buffer)) > 0 )
	{
		if (mp3file->id3.title == NULL)
			mp3file->id3.title = g_malloc0(MAX_STRING_LEN+1);
		if (buflen >= MAX_STRING_LEN)
			buflen = MAX_STRING_LEN;
		if (strncmp(buffer,mp3file->id3.title,buflen)!=0)
			/* Tag was changed, and must be saved */
			mp3file->changed |= TAG_CHANGED;
		strncpy(mp3file->id3.title,buffer,buflen+1);
	} else
	{
		if (mp3file->id3.title != NULL)
			g_free(mp3file->id3.title);
		mp3file->id3.title = NULL;
	}


	/**********
	 * Artist *
	 **********/
	buffer = gtk_entry_get_text(GTK_ENTRY(ArtistEntry));
	Id3tag_Strip(buffer);

	if ( (buflen=strlen(buffer)) > 0 )
	{
		if (mp3file->id3.artist == NULL)
			mp3file->id3.artist = g_malloc0(MAX_STRING_LEN+1);
		if (buflen >= MAX_STRING_LEN)
			buflen = MAX_STRING_LEN;
		if (strncmp(buffer,mp3file->id3.artist,buflen)!=0) 
			mp3file->changed |= TAG_CHANGED;
		strncpy(mp3file->id3.artist,buffer,buflen+1);
	} else
	{
		if (mp3file->id3.artist != NULL)
			g_free(mp3file->id3.artist);
		mp3file->id3.artist = NULL;
	}


	/*********
	 * Album *
	 *********/
	buffer = gtk_entry_get_text(GTK_ENTRY(AlbumEntry));
	Id3tag_Strip(buffer);

	if ( (buflen=strlen(buffer)) > 0 )
	{
		if (mp3file->id3.album == NULL)
			mp3file->id3.album = g_malloc0(MAX_STRING_LEN+1);
		if (buflen >= MAX_STRING_LEN)
			buflen = MAX_STRING_LEN;
		if (strncmp(buffer,mp3file->id3.album,buflen)!=0) 
			mp3file->changed |= TAG_CHANGED;
		strncpy(mp3file->id3.album,buffer,buflen+1);
	} else
	{
		if (mp3file->id3.album != NULL)
			g_free(mp3file->id3.album);
		mp3file->id3.album = NULL;
	}


	/********
	 * Year *
	 ********/
	buffer = gtk_entry_get_text(GTK_ENTRY(YearEntry));
	Id3tag_Strip(buffer);

	if ( (buflen=strlen(buffer)) > 0 )
	{
		if (mp3file->id3.year == NULL)
			mp3file->id3.year = g_malloc0(MAX_STRING_LEN+1);
		if (buflen >= MAX_STRING_LEN)
			buflen = MAX_STRING_LEN;
		if (strncmp(buffer,mp3file->id3.year,buflen)!=0) 
			mp3file->changed |= TAG_CHANGED;
		strncpy(mp3file->id3.year,buffer,buflen+1);
	} else
	{
		if (mp3file->id3.year != NULL)
			g_free(mp3file->id3.year);
		mp3file->id3.year = NULL;
	}


	/*********
	 * Track *
	 *********/
	buffer = g_strdup_printf("%.2d",atoi(gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(TrackEntry)->entry))));
	//Id3tag_Strip(buffer);

	if ( (buflen=strlen(buffer))>0 && atoi(buffer)!=0 )
	{
		if (mp3file->id3.track == NULL)
			mp3file->id3.track = g_malloc0(MAX_STRING_LEN+1);
		if (buflen >= MAX_STRING_LEN)
			buflen = MAX_STRING_LEN;
		if (strncmp(buffer,mp3file->id3.track,buflen)!=0) 
			mp3file->changed |= TAG_CHANGED;
		strncpy(mp3file->id3.track,buffer,buflen+1);
	} else
	{
		if (mp3file->id3.track != NULL)
			g_free(mp3file->id3.track);
		mp3file->id3.track = NULL;
	}
	g_free(buffer);


	/*********
	 * Genre *
	 *********/
	/* Create into buffer1 a format as '(genre_num)' */
	buffer = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(GenreEntry)->entry));
	Id3tag_Strip(buffer);
	buffer1 = g_strdup_printf("(%d)",Id3tag_String_To_Genre(buffer));

	if ( strlen(buffer) > 0 )
	{
		if (mp3file->id3.genre == NULL)
			mp3file->id3.genre = g_malloc0(MAX_STRING_LEN+1);
		if ((buflen=strlen(buffer1)) >= MAX_STRING_LEN)
			buflen = MAX_STRING_LEN;
		if (strncmp(buffer1,mp3file->id3.genre,buflen)!=0) 
			mp3file->changed |= TAG_CHANGED;
		strncpy(mp3file->id3.genre,buffer1,buflen+1);
	} else
	{
		/* If genre was '(255)' or is 'Unknown', we change nothing! */
		if ( mp3file->id3.genre!=NULL && strcmp(mp3file->id3.genre,"(255)")!=0
		     && strcasecmp(buffer,"Unknown")!=0 )
		{
			//if (mp3file->id3.genre != NULL)
				g_free(mp3file->id3.genre);
			mp3file->id3.genre = NULL;
		}
	}
	g_free(buffer1);


	/***********
	 * Comment *
	 ***********/
	buffer = gtk_entry_get_text(GTK_ENTRY(CommentEntry));
	Id3tag_Strip(buffer);

	if ( (buflen=strlen(buffer)) > 0 )
	{
		if (mp3file->id3.comment == NULL)
			mp3file->id3.comment = g_malloc0(MAX_STRING_LEN+1);
		if (buflen >= MAX_STRING_LEN)
			buflen = MAX_STRING_LEN;
		if (strncmp(buffer,mp3file->id3.comment,buflen)!=0) 
			mp3file->changed |= TAG_CHANGED;
		strncpy(mp3file->id3.comment,buffer,buflen+1);
	} else
	{
		if (mp3file->id3.comment != NULL)
			g_free(mp3file->id3.comment);
		mp3file->id3.comment = NULL;
	}

	/* Put data in undo list if needed */
	Undo_Add_Data(mp3file);
}



/*
 * Returns the corresponding genre value of the input string,
 * else returns 0xFF(unknown genre, but not invalid).
 */
guchar Id3tag_String_To_Genre (gchar *genre)
{
	gint i;

	for (i=0; i<GENRE_MAX; i++)
		if (strcasecmp(genre,id3_genres[i])==0)
			return (guchar)i;
	return (guchar)0xFF;
}


/*
 * Returns the name of a genre code if found
 * Three states for genre code :
 *    - defined (0 to GENRE_MAX)
 *    - undefined/unknown (GENRE_MAX+1 to ID3_INVALID_GENRE-1)
 *    - invalid (>ID3_INVALID_GENRE)
 */
gchar *Id3tag_Genre_To_String (unsigned char genre_code)
{
	if (genre_code>=ID3_INVALID_GENRE)    /* empty */
		return "";
	else if (genre_code>GENRE_MAX)        /* unknown tag */
		return "Unknown";
	else                                  /* known tag */
		return id3_genres[genre_code];
}


/*
 * Load the track numbers (30 minimum) into the track combo list
 */
void Id3tag_Load_Track_List (void)
{
	GList *tracklist = NULL;
	gint len;
	gint i;

	if (!Mp3FileList || !TrackEntry) return;

	if ((len=g_list_length(Mp3FileList)) < 30)
		len = 30;
	
	/* Create GList */
	if (!tracklist)
	{
		tracklist = g_list_append(tracklist," ");
		for (i=1; i<=len; i++)
			tracklist = g_list_append(tracklist,g_strdup_printf("%.2d",i));
	}

	/* Connect the combo and the glist */
	gtk_combo_set_popdown_strings(GTK_COMBO(TrackEntry),tracklist);
}


/*
 * Delete spaces at the end and the beginning of the string 
 */
void Id3tag_Strip (gchar *string)
{
	if (!string) return;
	string = g_strstrip(string);
}



/*
 * As the ID3Field_GetASCII function differs with some versions of id3lib,
 * we must redefine it.
 */
ID3_C_EXPORT size_t my_ID3Field_GetASCII(const ID3Field *field, char *buffer, size_t maxChars, index_t itemNum)
{
#    if (ID3LIB_MAJOR_VERSION >= 3)
#        if (ID3LIB_MINOR_VERSION <= 7)
//#            if (ID3LIB_PATCH_VERSION <= 12)
                 // (<= to 3.7.12)
                 return ID3Field_GetASCII(field,buffer,MAX_STRING_LEN,itemNum);
//                 return ID3Field_GetASCII(field,buffer,MAX_STRING_LEN,1);
//#            else
                 // (>= to 3.7.13)
//                 return ID3Field_GetASCII(field,buffer,MAX_STRING_LEN,0);
//#            endif
#        elif (ID3LIB_MINOR_VERSION >= 8)
             // (>= to 3.8.x)
             return ID3Field_GetASCII(field,buffer,MAX_STRING_LEN);
//#        else
             // Not tested (< 3.7.x)
//             return ID3Field_GetASCII(field,buffer,MAX_STRING_LEN,1);
#        endif
#    else
         // Not tested (< 3.x.x)
         return ID3Field_GetASCII(field,buffer,MAX_STRING_LEN,itemNum);
#    endif
}
